/*
 *	chasm.c - main module
 *
 *	@(#)chasm.c	1.1 91/04/10
 *
 *	Copyright (c) 1991 Steve Scherf
 *
 *	Author:	Steve Scherf
 *	Date:	Wed Apr 10 22:53:11 PDT 1991
 *----------------------------------------------------------------------
 *	Amiga porting:
 *	
 *	By:	Gian Uberto Lauri
 *	Date: Wed Jun 19 1991
 *
 * Notes:
 *
 *	This code made a fine usage of the Unix system calls to link and
 *	remove, not available in AmigaDOS. I switched to standard library
 *	rename and remove.
 *
 *	It required an extra fclose, too.
 *
 *	Since unlink was changed in remove with a string search & replace,
 *	sunlink(...) was changed in sremove(...) too.
 *	The command line parsing was made more readable.
 *	This file is for use with Amiga and MS-DOS
 *----------------------------------------------------------------------
 *	SUPERCHIPextension:
 *	
 *	By:	Gian Uberto Lauri
 *	Date: Mon Jun 24 1991
 *
 * Notes:
 *	I added the new tokens last, without care of their alphabetic
 *	order. I choosed 3 letters mnemonics for sake of consistency
 *	with the old code, and in most case they are acronyms.
 *
 *	ESA	Extended Screen Activation (Sorry, On wasn't a good choice)
 *	ESO	Extended Screen Off
 *	BYE	BYE (leave superchip - that's not an acronym :) )
 *	SDP	Set Digit Pointer
 *	WUF	Write User Flags (shut up doggie!)
 *	RUF	Read User Flags
 *	SCD	Scroll down n pixel
 *	SCR	Scroll right 4 pixel
 *	SCL	Scroll left 4 pixel
 *
 *	For more details about S-CHIP, see the relative posting.
 *	BTW, adding new instructions using bison/yacc is trivial.*
 */

#include <stdio.h>
#include <string.h>
#include "chasm.h"

char *whatstr = "@(#) chasm v0.1	Copyright (c) 1991 Steve Scherf";

int aflag;			/* true if ascii output */
int oflag;			/* true if redfined output filename */

char *infile;
char *outfile;
char *cout	= "c.out";
char *cascii	= "c.ascii";
char *ctmp	= "c.tmp";
char *xstrrchr();

FILE *ifp;
FILE *ofp;


main(argc, argv)
int argc;
char *argv[];
{
	int i;
	char *name, *c;

	name = argv[0];
	infile = argv[argc-1];

	if(!--argc)
		usage(name);

	for(i = 1 ; i < argc; i++) {
		if(*argv[i] != '-')
			usage(name);
		c = argv[i];
		while(*(++c))
			switch(*c) {
				case 'a':
					aflag++;
					break;
				case 'o':
					if(!(outfile = argv[++i]))
						usage(name);
					oflag++;
					break;
				default:
					usage(name);
			}
	}

  	if(!oflag)
		if(aflag)
			outfile = cascii;
		else
			outfile = cout;

	if((ifp = fopen(infile, "r")) == 0) {
		fprintf(stderr, "%s: cannot open %s\n", name, infile);
		exit(1);
	}

	if((ofp = fopen(ctmp, "w")) == 0) {
		fprintf(stderr, "%s: cannot open %s\n", name, ctmp);
		exit(1);
	}

	init_links();
	yyparse();
	chk_var();

	if(chk_size() || (chk_val() + chk_lab()) || pars_err ||
	    !out_inst(ofp, aflag)) {
		if(remove(ctmp))
			fprintf(stderr,"%s: can't remove %s",name,ctmp);
		printf("No output written to %s\n", outfile);
		exit(1);
	}

	if(sremove(name,outfile))
		exit(1);

	fclose(ofp);
	if(rename(ctmp, outfile)) {
		fprintf(stderr, "Cannot rename %s to %s\n", ctmp, outfile);
		sremove(name,ctmp);
		exit(1);
	}

	sremove(name,ctmp);
}


/* remove only if file exists */
sremove(fn,n)
char *fn,*n;
{
	if(!access(n, 0) && remove(n)) {
		fprintf(stderr,"%s: can't remove %s",fn,n);
		return -1;
	}

	return 0;
}


/* print usage message and exit */
usage(name)
char *name;
{
	char *n;

	if(n = xstrrchr(name, '/'))
		n++;
	else
		n = name;

	fprintf(stderr, "Usage: %s [-a] [-o output_file] input_file\n", n);
	exit(1);
}


/* return a pointer to the char c in s, null if nonexistent */
char *
xstrrchr(s, c)
register char *s;
register char c;
{
	register char *p = 0;

	while(*s) {
		if(*s == c)
			p = s;
		s++;
	}

	return p;
}
